/*******************************************************************************
*                                                                              *
* System      : SiTCP                                                          *
* Block       : SIO                                                            *
* Module      : SIO_MASTER                                                     *
* Version     : v 0.0.0 2008/03/26 12:01                                       *
*                                                                              *
* Description : Serial IO interface (Master)                                   *
*                                                                              *
* Designer    : Tomohisa Uchida                                                *
*                                                                              *
*                Copyright (c) 2008 Tomohisa Uchida                            *
*                All rights reserved                                           *
*                                                                              *
*******************************************************************************/
module SIO_MASTER(
	// System
	CLK					,	// in	: System clock
	RST					,	// in	: System reset
	// Timer
	TIM_1US				,	// in	: 1us interrupt
	ADDR_WIDTH			,	// in	: Address width[1:0]
							//		   0: 8bit, 1:16bit, 2:24bit, 3:32bit
	// SiTCP I/F
	LOC_ACT				,	// in	: Request
	LOC_ADDR			,	// in	: Address[31:0]
	LOC_WD				,	// in	: Data[7:0]
	LOC_WE				,	// in	: Write enable
	LOC_RE				,	// in	: Read enable
	LOC_ACK				,	// out	: Access acknowledge
	LOC_RD				,	// out	: Read data[7:0]
	// Serial I/F
	SCK					,	// out	: Clock
	SCS					,	// out	: Active
	SI					,	// in	: Data input
	SO						// out	: Data output
);

//-------- Input/Output -------------
	input			CLK				;
	input			RST				;

	input			TIM_1US			;
	input	[1:0]	ADDR_WIDTH		;

	input			LOC_ACT			;
	input	[31:0]	LOC_ADDR		;
	input	[7:0]	LOC_WD			;
	input			LOC_WE			;
	input			LOC_RE			;
	output			LOC_ACK			;
	output	[7:0]	LOC_RD			;

	output			SCK				;
	output			SCS				;
	input			SI				;
	output			SO				;

//---------- Buffer ----------
	wire			LOC_ACK			;
	wire	[7:0]	LOC_RD			;

	wire			SCK				;
	wire			SCS				;
	wire			SO				;

//------------------------------------------------------------------------------
//	
//------------------------------------------------------------------------------
//--------------------------------------
//	Input registers
//--------------------------------------
	reg				irTim1us		;

	always@ (posedge CLK) begin
		irTim1us		<= TIM_1US;
	end

	reg		[1:0]	irAddrWidth		;
	reg				irLocAct		;
	reg		[31:0]	irLocAddr		;
	reg		[7:0]	irLocWd			;
	reg				irLocWe			;
	reg				irLocRe			;
	reg				irLocLd			;
	reg				irSi			;

	always@ (posedge CLK) begin
		irAddrWidth[1:0]	<= ADDR_WIDTH[1:0];

		irLocAct		<= LOC_ACT;
		irLocAddr[31:0]	<= LOC_ADDR[31:0];
		irLocWd[7:0]	<= LOC_WD[7:0];
		irLocWe			<= LOC_WE;
		irLocRe			<= LOC_RE;
		irLocLd			<= LOC_WE | LOC_RE;

		irSi			<= SI;
	end

//--------------------------------------
// I/F error detection
//--------------------------------------
//      Inhibit LOC_ACK to generate LOC_ERR

	reg				ifErrDet		;

	always@ (posedge CLK or posedge RST) begin
		if(RST)begin
			ifErrDet	<= 1'b0;
		end else begin
			ifErrDet	<= irSi & (irLocLd | ifErrDet);
		end
	end

//------------------------------------------------------------------------------
//	Send
//------------------------------------------------------------------------------
//--------------------------------------
//	
//--------------------------------------
	reg		[7:0]	cmdReg			;
	reg		[39:0]	dataReg			;

	always@ (posedge CLK) begin
		if(irLocLd)begin
			cmdReg[7:0]		<= {1'b1,1'b1,irLocRe,1'b0,irAddrWidth[1:0],2'd0};

			case(irAddrWidth[1:0])
				2'd0: begin
					dataReg[39:0]	<= {irLocAddr[ 7:0],irLocWd[7:0],24'hFFFFFF};
				end
				2'd1: begin
					dataReg[39:0]	<= {irLocAddr[15:0],irLocWd[7:0],16'hFFFF};
				end
				2'd2: begin
					dataReg[39:0]	<= {irLocAddr[23:0],irLocWd[7:0],8'hFF};
				end
				default: begin
					dataReg[39:0]	<= {irLocAddr[31:0],irLocWd[7:0]};
				end
			endcase
		end
	end

	reg				sendReq			;
	reg				sendAccept		;

	always@ (posedge CLK) begin
		sendReq	<= irLocAct & ~sendAccept & (irLocLd | sendReq);
	end


//--------------------------------------
//	Clock generator
//--------------------------------------
	reg				sclk			;
	reg				dlySclk			;
	reg				trgSck			;

	always@ (posedge CLK or posedge RST) begin
		if(RST)begin
			sclk	<= 1'b0;
			dlySclk	<= 1'b0;
			trgSck	<= 1'b0;
		end else begin
			if(irTim1us)begin
				sclk	<= ~sclk;
			end
			dlySclk	<=  sclk;
			trgSck	<=  dlySclk & ~sclk;
		end
	end

//--------------------------------------
//	Send data buffer
//--------------------------------------
	always@ (posedge CLK) begin
		sendAccept	<= irLocAct & trgSck & sendReq;
	end

	reg				sendAct			;
	reg		[48:0]	sendBuf			;
	reg				sendSck			;
	reg				recdAck			;

	reg				ackBuf			;

	always@ (posedge CLK) begin
		if(trgSck)begin
			sendAct			<= irLocAct & ~ackBuf & (sendReq | sendAct);
			sendBuf[48:0]	<= (sendReq	? {1'b0,cmdReg[7:0],dataReg[39:0]}	: {sendBuf[47:0],1'b1});
		end

		sendSck	<= dlySclk;
	end

	reg				orScs			;
	reg				orSck			;
	reg				orSo			;

	always@ (posedge CLK) begin
		orScs	<= sendAct;
		orSck	<= sendSck;
		orSo	<= sendAct & sendBuf[48];
	end

	assign	SCS		= orScs;
	assign	SCK		= orSck;
	assign	SO		= orSo;

//------------------------------------------------------------------------------
//	Receive
//------------------------------------------------------------------------------
	reg		[7:0]	recvBuf			;

	always@ (posedge CLK) begin
		if(~sendAct)begin
			recvBuf[7:0]	<= 8'd0;
			ackBuf			<= 1'b0;
		end else begin
			if(trgSck)begin
				recvBuf[7:0]	<= (sendBuf[39] ? {recvBuf[6:0],irSi} : 8'd0);
				ackBuf			<= (cmdReg[5]	? recvBuf[7] : irSi);
			end
		end
	end

	reg				inhLocAck		;
	reg				orLocAck		;
	reg		[7:0]	orLocRd			;

	always@ (posedge CLK) begin
		inhLocAck		<=  sendAct & (trgSck & ackBuf | inhLocAck);
		orLocAck		<=  ~inhLocAck & trgSck & ackBuf;
		orLocRd[7:0]	<= (~inhLocAck & trgSck & ackBuf ? recvBuf[7:0] : 8'd0);
	end

	assign	LOC_ACK		= orLocAck;
	assign	LOC_RD[7:0]	= orLocRd[7:0];

//------------------------------------------------------------------------------
endmodule
